
/*
 *  matfile.c
 *  match files
 *
 *  @author Steve Hoffmann
 *  @email steve@bioinf.uni-leipzig.de
 *  @date 08/25/2010 03:42:37 PM CEST
 *  
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include "alignment.h"
#include "stringutils.h"
#include "basic-types.h"
#include "mathematics.h"
#include "bitVector.h"
#include "matchfiles.h"
#include "browsematchfiles.h"
#include "matfile.h"
#include "iupac.h"
#include "info.h"
#include "manopt.h"
#include "evalmatchfiles.h"
#include "evalmethylmatchfiles.h"
#include "splicesites.h"
#include "startsites.h"

unsigned char mute = 0;
char *ntcode;


/*---------------------------------- stats -----------------------------------
 *    
 * @brief get matchfile stats
 * @author Steve Hoffmann 
 *   
 */

  void
getstats (void *space, matchfile_t *file, fasta_t *fasta, Uint mincover, Uint maxcover, double minfrac, char entropyfilter, char GEV, char usenativequal)
{

  double ws[]={0.95, 0.05};
  Uint ret,i,u;
  double mx =.0 , sx = .0, kx = .0;
  matchfileSampleStats_t *stats;

  MSG("evaluating statistics\n");
  stats = bl_matchfileInitSampleStats(space, 120000, mincover, maxcover, minfrac, entropyfilter, 10, .15);
  //bl_matchfileDumpFileStats(space, file);
  //bl_matchfileEvalCrossSections(space, file, fasta, NULL, NULL);
  stats->usenativequal = usenativequal;
  file->index->stats = stats;

  MSG("evaluating error density.\n");
  ret = bl_matchfileSampleCrossSections(space, file, fasta, 20000, 
      bl_matchfileGetErrorDensity, stats);


  if(stats->entropydensity == NULL) {

    stats->entropydensitystep=0.05;
    stats->entropydensitylen = (Uint)(2.0/stats->entropydensitystep) + 1;

    stats->entropydensity =
      ALLOCMEMORY(space, NULL, double, stats->entropydensitylen+1);

    memset(stats->entropydensity, 0, sizeof(double)*(stats->entropydensitylen+1));

    for(i=0; i < stats->e_N; i++) {
      fprintf(stderr, "entopy[%d]=%.3f (", i, stats->entropy[i]);
      for(u=1; u < stats->entropydensitylen; u++) {
        if(stats->entropy[i] <= ((double)u)*0.05) {
          fprintf(stderr,"%.3f:%.3f,", ((double)u*0.05),stats->entropydensity[u]);
          stats->entropydensity[u]+=1.0;
        }
      }
      fprintf(stderr, "\n");
    }

    for(u=1; u < stats->entropydensitylen; u++) {
      fprintf(stderr, "raw %d: %.3f\n", u,
          stats->entropydensity[u]);
      stats->entropydensity[u]/=(double)stats->e_N;
      fprintf(stderr, "norm %d: %.3f\n", u,
          stats->entropydensity[u]);
    }   
  }


  if(!ret) return;

  stats->e_ll=gmm(space, stats->e, stats->e_N, 1, stats->e_mu, 
      stats->e_sd, ws, 2, 10000);

  fprintf(stderr, "mu*[]={ %f %f }, sd*[]={ %f %f }, ws={ %f, %f}, ll=%f\n", 
      stats->e_mu[0], stats->e_mu[1], stats->e_sd[0], stats->e_sd[1], ws[0], ws[1], stats->e_ll);

  ws[0]=0.95;
  ws[1]=0.05;

  stats->b_ll=gmm(space, stats->b, stats->e_N, 1, &stats->b_mu, 
      &stats->b_sd, ws, 1, 10000);

  fprintf(stderr, "mu*[]={ %f }, sd*[]={ %f }, ll=%f\n", 
      stats->b_mu, stats->b_sd, stats->b_ll);


  if(!GEV ) { 

    stats->px =stats->e_mu[0] + stats->e_sd[0];
    while(univarnormcdf(stats->px, stats->e_mu[0], stats->e_sd[0]) < 0.99) 
      stats->px += 0.00001;
  
  } else { 

    qsort(stats->eraw, stats->e_N, sizeof(double), cmp_dbl_qsort);
    gevLmoment(stats->eraw, stats->e_N, &mx, &sx, &kx);

    NFO("lmoment m:%f, s:%f, xi:%f \n", mx, sx, -kx);
    NFO("log-likelihood:%f\n", gevll(stats->eraw, stats->e_N, mx, sx, kx));

    gevmle(NULL, stats->eraw, stats->e_N, &mx, &sx, &kx, 10000, stats->eraw[0], stats->eraw[stats->e_N-1]);
    NFO("gev m:%f, s:%f, xi:%f \n", mx, sx, -kx);

    stats->px = mx;
    while(gevcdf(stats->px, mx, sx, -kx) < 0.99) 
      stats->px += 0.00001;

    stats->pxx = mx;
    while(gevcdf(stats->px, mx, sx, -kx) < 0.66) 
      stats->pxx += 0.00001;


  }

  NFO("99%% quantile is px: %f\n", stats->px);
  NFO("66%% quantile is px: %f\n", stats->pxx);
 
  MSG("sample cross sections.\n");
  ret = bl_matchfileSampleCrossSections(space, file, fasta, 100000, 
      bl_matchfileGetConditionals, stats);

//  if(!ret) return;


  ws[0]=0.95;
  ws[1]=0.05;

  stats->V_ll=gmm(space, stats->V, stats->V_N, 1, 
      &stats->V_mu, &stats->V_sd, ws, 1, 100000); 
 
  NFO("fitted readstart variance with gaussian: mu=%f, sd=%f, ll=%f\n", 
      stats->V_mu, stats->V_sd, stats->V_ll);


  stats->Vx_ll=gmm(space, stats->Vx, stats->Vx_N, 1, 
      &stats->Vx_mu, &stats->Vx_sd, ws, 1, 100000);


  return ;
}

/*----------------------------------- view -----------------------------------
 *    
 * @brief view the matchfiles
 * @author Steve Hoffmann 
 *   
 */

void
view (void *space, matchfile_t **files, Uint nooffiles, fasta_t *fasta, 
    annotationtrack_t *bed)
{
 
  MSG("starting viewer.\n");
  bl_matchfileViewer(space, files, nooffiles, fasta, bed, 100, 1000);
  return ;
}

/*----------------------------------- eval -----------------------------------
 *    
 * @brief evaluation of matchfiles
 * @author Steve Hoffmann 
 *   
 */
 
void
eval (void *space, matchfile_t **files, int *groups, Uint nooffiles, fasta_t *fasta, Uint maxframesize)
{

  
  MSG("evaluating x-sections.\n");
  
  bl_matchfileEvalCrossSections(space, files, groups, nooffiles, fasta, maxframesize, 
      bl_matchfileTest, NULL);
 
  return ;
}
/*----------------------------------- simpleeval -----------------------------------
 *    
 * @brief simple evaluation of matchfiles
 * @author Steve Hoffmann 
 *   
 */
 
void
simpleeval (void *space, matchfile_t **files, int *groups, Uint nooffiles, fasta_t *fasta, Uint maxframesize, char usenativequal)
{

  
  MSG("simple evaluating x-sections.\n");
  
  bl_matchfileEvalCrossSections(space, files, groups, nooffiles, fasta, maxframesize, 
      bl_matchfileSimpleGEV, &usenativequal);
 
  return ;
}

/*----------------------------------- gatkeval -----------------------------------
 *    
 * @brief simple evaluation of matchfiles
 * @author Steve Hoffmann 
 *   
 */
 
void
gatkeval (void *space, matchfile_t **files, int *groups, Uint nooffiles, fasta_t *fasta, Uint maxframesize, char usenativequal)
{

  
  MSG("gatk evaluating x-sections.\n");
  
  bl_matchfileEvalCrossSections(space, files, groups, nooffiles, fasta, maxframesize, 
       bl_matchfileSimpleGATK , &usenativequal);
 
  return ;
}

/*---------------------------------- consensus ---------------------------------
 *    
 * @brief simple evaluation of matchfiles
 * @author Christian Otto
 *   
 */

void
evalconsensus (void *space, matchfile_t **files, int *groups, Uint nooffiles, fasta_t *fasta, Uint maxframesize)
{

  
  MSG("consensus evaluating x-sections.\n");
  
  bl_matchfileEvalCrossSections(space, files, groups, nooffiles, fasta, maxframesize, 
       bl_matchfileConsensus , files);
 
  return ;
}

/*------------------------------ callMethylSimple ------------------------------
 *    
 * @brief simple evaluation of matchfiles
 * @author Christian Otto
 *   
 */

void
callMethylSimple (void *space, matchfile_t **files, int *groups, Uint nooffiles, fasta_t *fasta, Uint maxframesize)
{
  assert(nooffiles == 1);

  Uint i, len;
  char *name, *basename;
  matfile_t matfile;
  matfile.dev = stdout;
  matfile.fasta = fasta;
  matfile.files = files;  
  
  /* prepare sample name from file name */
  basename = bl_basename(files[0]->filename);
  len = strlen(basename);
  name = ALLOCMEMORY(space, NULL, char, len + 1);
  memmove(name, basename, len + 1);

  if (files[0]->gzip){
    if (strncmp(&name[len-3], ".gz", 3) == 0){
      name[len-3] = '\0';
    }
    else {
      if (strncmp(&name[len-5], ".gzip", 5) == 0) {
        name[len-5] = '\0';
      }
    }  
    assert(strlen(name) != len);
    len = strlen(name);
  }
  len = bl_fileprefixlen(name);
  name[len] = '\0';
  
  MSG("evaluating x-sections.\n");
  
  /* output VCF header */
  fprintf(matfile.dev, "##fileformat=VCFv4.1\n");
  
  for (i = 0; i < fasta->nooffiles; i++){
    fprintf(matfile.dev, "##reference=file://%s\n",
            fasta->filenames[i]);
  }

  for (i = 0; i < fasta->noofseqs; i++){
    fprintf(matfile.dev, "##contig=<ID=%s,length=%d>\n", 
            bl_fastaGetDescription(fasta, i), 
            bl_fastaGetSequenceLength(fasta,i));
  }
  //fileDate? program call?
  fprintf(matfile.dev, "##INFO=<ID=CS,Number=1,Type=Character,Description=\"Bisulfite conversion strand, i.e., strand of cytosine relative to reference\">\n");
  fprintf(matfile.dev, "##INFO=<ID=CC,Number=1,Type=String,Description=\"Sequence context of two bases on the reference relative to bisulfite conversion strand, missing value if context is beyond reference boundaries\">\n");
  fprintf(matfile.dev, "##INFO=<ID=NS,Number=1,Type=Integer,Description=\"Number of samples with data\">\n");
  fprintf(matfile.dev, "##INFO=<ID=MMR,Number=1,Type=Float,Description=\"Mean methylation rate among samples with data\">\n");
  fprintf(matfile.dev, "##INFO=<ID=DMR,Number=1,Type=Float,Description=\"Difference in methylation rate (Sample1 - Sample2).\">\n");
  fprintf(matfile.dev, "##FORMAT=<ID=DP,Number=1,Type=Integer,Description=\"Total Depth, not filtered by mapping quality, base quality, or bisulfite conversion information\">\n");
  fprintf(matfile.dev, "##FORMAT=<ID=MDP,Number=1,Type=Integer,Description=\"Number of reads after filtering by bisulfite conversion strand\">\n");
  fprintf(matfile.dev, "##FORMAT=<ID=MDP3,Number=3,Type=Integer,Description=\"Number of reads with 1) unconverted cytosines, 2) converted cytosines or thymines, and 3) other bases after filtering by bisulfite conversion strand\">\n");
  fprintf(matfile.dev, "##FORMAT=<ID=MRDP,Number=1,Type=Integer,Description=\"Number of reads with unconverted (methylated) and converted (unmethylated) cytosines after filtering by bisulfite conversion strand\">\n");
  fprintf(matfile.dev, "##FORMAT=<ID=CM,Number=1,Type=Integer,Description=\"Number of reads with unconverted (methylated) cytosines after filtering by bisulfite conversion strand\">\n");
  fprintf(matfile.dev, "##FORMAT=<ID=CU,Number=1,Type=Integer,Description=\"Number of reads with converted (unmethylated) cytosines after filtering by bisulfite conversion strand\">\n");
  fprintf(matfile.dev, "##FORMAT=<ID=MR,Number=1,Type=Float,Description=\"Estimated methylation rate\">\n");
  fprintf(matfile.dev, "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t%s\n", name);

  bl_matchfileEvalCrossSections(space, files, groups, nooffiles, fasta, maxframesize, 
       bl_matchfileCallMethylSimple, &matfile);

  FREEMEMORY(space, name); 
  return ;
}

void
calcMethylBias (void *space, matchfile_t **files, int *groups, Uint nooffiles, fasta_t *fasta, Uint maxframesize)
{
  matfile_t matfile;
  matfile.dev = stdout;
  matfile.fasta = fasta;
  matfile.files = files;

  bl_matchfileEvalCrossSections(space, files, groups, nooffiles, fasta, maxframesize, 
       bl_matchfileCalcMethylBias, &matfile);
 
  return ;
}


/*-------------------------------- evalstarts --------------------------------
 *    
 * @brief get start sites
 * @author Steve Hoffmann 
 *   
 */
 
void
evalstarts (void *space, matchfile_t **files, int *groups, Uint nooffiles, 
    fasta_t *fasta, Uint maxframesize, annotationtrack_t *bed )
{
 
  Uint *cntr;
//  Uint i;
  cntr = ALLOCMEMORY(space, NULL, Uint, 255);
  memset(cntr, 0, sizeof(Uint)*255);
 
  MSG("start sites.\n");
/*
  bl_matchfileEvalCrossSections(space, files, groups, nooffiles, fasta, 
      bl_matchfileStartSites, cntr);

  for(i=0; i < 255; i++) {
    printf("%d\t%d\n", i, cntr[i]);
  }
*/
  
  MSG("coverage.\n");
  bl_matchfileEvalCrossSections(space, files, groups, nooffiles, fasta, maxframesize, 
      bl_coverage, cntr);

  return ;
}

/*-------------------------------- writewiggle -------------------------------
 *    
 * @brief get start sites
 * @author Steve Hoffmann 
 *   
 */
 
void
writeexpression (void *space, matchfile_t **files, int *groups, Uint nooffiles, 
    fasta_t *fasta, annotationtrack_t *bed)
{
 
  Uint i;
  char *filename;
  MSG("write expression file.\n");

  for(i=0; i < nooffiles; i++) { 
    filename = ALLOCMEMORY(space, NULL, char, strlen(files[i]->filename)+5);
    memmove(filename, files[i]->filename, strlen(files[i]->filename));
    sprintf(&filename[strlen(files[i]->filename)], ".wig");

    bl_writeexpression(space, filename, strlen(filename), files[i], fasta, 
        10000000, 1000000);
  }

  return ;
}



/*---------------------------------- splice ----------------------------------
 *    
 * @brief dump the splice sites
 * @author Steve Hoffmann 
 *   
 */
 
void
evalsplice (void *space, matchfile_t **files, int *groups, Uint nooffiles, fasta_t *fasta, Uint maxframesize, 
    annotationtrack_t *bed, char *filename, char *transfilename, Uint minsplitno)
{
  Uint i, cidx, pos, len;
  char *chr, *chr2, *ptr, *base;
  splitmap_t map;
  splicemap_t *sm;
  FILE *fp1, *fp2;

  bl_matchfileInitSplitMap(space, &map, bed, files, fasta);
  MSG("eval splice sites.\n");
  bl_matchfileEvalCrossSections(space, files, groups, nooffiles, fasta, maxframesize, 
      bl_matchfileSplit, &map);
  MSG("condensing sites.\n");
  sm = bl_matchfileSpliceMap (space, &map, 10, minsplitno);
  MSG("writing splice sites to stdout.\n");



  fp1 = fopen(filename, "w");
  if (fp1 == NULL) {
    fprintf(stderr, "Couldnt open %s for reading. Exit forced.\n", filename);
    exit(-1);
  } 


  fp2 = fopen(transfilename, "w");
  if (fp2 == NULL) {
    fprintf(stderr, "Couldnt open %s for reading. Exit forced.\n", transfilename);
    exit(-1);
  } 


  base = bl_basename(filename);

  //printsplice(space, sm, stdout);
  printsplicebed(space, sm, minsplitno, base, fp1, fp2);
 
  MSG("writing splice sites to gff.\n");
  if(bed) {
    bl_annotationtrackGetStats (space, bed);
    bl_matchfileSpliceAnnotation(space, sm, bed);
    bl_GFFwrite("splice.gff", bed);
  }

  for(i=0; i < map.noofsplits; i++) {
    cidx = map.cidx[i];
    chr = bl_fastaGetDescription(fasta, cidx);
    pos = map.pos[i];
 
    len = strlen(chr);
    chr2 = ALLOCMEMORY(space, NULL, char, len+1);
    memmove(chr2, chr, len);
    chr2[len] = 0;
    ptr = strtok(chr2, " ");
    
    // dummy if to avoid not used warnings
    if (0){
      printsplits(space, ptr, pos, &map.cs[i], &map);
    }
    FREEMEMORY(space, chr2);
  }

  bl_matchfileDestructSplitMap(space, &map);
  bl_matchfileDestructSpliceMap (space, sm);  
  FREEMEMORY(space, sm);

  fclose(fp1);
  fclose(fp2);

  return ;
}

/*----------------------------------- main -----------------------------------
 *    
 * @brief the main
 * @author Steve Hoffmann 
 *   
 */
 

int
main(int argc, char **argv) {
  
  void *space = NULL;
  
  manopt_optionset optset;
  manopt_arg *unflagged; 
  manopt_arg *dbfilenames;
  manopt_arg *queries;
  manopt_arg *indices;
  manopt_arg *grouplist;
  manopt_arg *bedfilenames;
  matchfile_t **files = NULL;
  annotationtrack_t *track = NULL;
  fasta_t *fasta = NULL;
  matchfileindex_t *idx2 = NULL;
  Uint prefixlen=0; 
  Uint splicebasenamelen;
  Uint minsplitno = 4;
  Uint mincover = 6;
  Uint maxcover = 200;
  char entropyfilter = 0;
  double minfrac = 0.1;
  unsigned char gzip = 0, browse=0, call=0, saveindex=0, stats=0, dumpstats=0, starts=0, wiggle=0, GEV=0, consensus=0, methylcall=0, methylbias=0;
  char version[]="0.1";
  char *splicebasename = NULL;
  char usenativequal = 0, simplegev=0, gatk=0;
  int *groups = NULL;
  int i;
  Uint maxframesize = 100000;

  FILE *beddev = NULL;
  char *filename;
  char *transfilename;
   
  initIUPAC(1,1); 
  manopt_initoptionset(&optset, argv[0], NULL, 
      "Heuristic mapping of short sequences\n",
      "SEGEMEHL is free software for non-commercial use \n  (C) 2008 Bioinformatik Leipzig\n",
      version,
      "Please report bugs to steve@bioinf.uni-leipzig.de"); 
  manopt(&optset, LISTOPT, 1, 'd', "database", 
      "list of path/filename(s) of database sequence(s)", "<file> [<file> ...]", 
      NULL, NULL);
  manopt(&optset, LISTOPT, 1, 'q', "query", 
      "path/filename of alignment file", "<file> [<file> ...]", NULL, NULL); 
  manopt(&optset, LISTOPT, 0, 'i', "index", 
      "path/filename of db index", "[<file> ... ]", NULL, NULL);
  manopt(&optset, LISTOPT, 0, 'a', "annotation", 
      "path/filename of bed annotation", "[<bedfile> ... ]", NULL, NULL);
  manopt(&optset, LISTOPT, 0, 'x', "generate", 
      "generate db index and store to disk", "<file>", NULL, NULL); 
  manopt(&optset, LISTOPT, 0, 'g', "group", 
      "group number for all nput files (1-offset and at most #files groups)", "[<int> ...]", NULL, NULL); 
  manopt(&optset, REQSTRINGOPT, 0, 's', "splice", 
      "dump splice sites to <basename>", NULL, NULL, &splicebasename);
  manopt(&optset, FLAG, 0, 'S', "starts", 
      "dump start sites", NULL, NULL, &starts);
  manopt(&optset, FLAG, 0, 'b', "browse", 
      "start browser", NULL, NULL, &browse);
  manopt(&optset, FLAG, 0, 'c', "call",
      "variant caller", NULL, NULL, &call);
  manopt(&optset, FLAG, 0, 'w', "expression",
      "generate a expression graph file from the match files", NULL, NULL, &wiggle);
  manopt(&optset, FLAG, 0, 'Z', "dumpstats",
      "dump data stats", NULL, NULL, &dumpstats);
  manopt(&optset, FLAG, 0, 'H', "stats", 
      "stats calculation", NULL, NULL, &stats);
  manopt(&optset, REQUINTOPT, 0, 'm', "minsplitno", 
      "minimum number of splits required to call a splice site", "<n>",
      NULL, &minsplitno);
  manopt(&optset, REQUINTOPT, 0, 'N', "mincover", 
      "minimum coverage to call sites", "<n>",
      NULL, &mincover);
  manopt(&optset, REQUINTOPT, 0, 'M', "maxcover", 
      "minimum coverage to call sites", "<n>",
      NULL, &maxcover);
  manopt(&optset, REQUINTOPT, 0, 'K', "maxframesize", 
      "maximum size of memory resident frame", "<n>",
      NULL, &maxframesize);
  manopt(&optset, REQDBLOPT, 0, 'F', "minfrac", 
      "minimum fraction of ALT alleles to call sites", "<f>",
      NULL, &minfrac);
  manopt(&optset, FLAG, 0, 'E', "entropyfilter",
      "turn on entropy filter", NULL, NULL, &entropyfilter);
  manopt(&optset, FLAG, 0, 'G', "GEV",
      "use GEV", NULL, NULL, &GEV);
  manopt(&optset, FLAG, 0, 'Q', "nativequal",
      "use native quals", NULL, NULL, &usenativequal);
  manopt(&optset, FLAG, 0, 'V', "simplegev",
      "use simple gev", NULL, NULL, &simplegev);
  manopt(&optset, FLAG, 0, 'A', "gatk", 
      "use GATK model", NULL, NULL, &gatk);
  manopt(&optset, FLAG, 0, 'C', "consensus", 
      "use simple consensus calling", NULL, NULL, &consensus);
  manopt(&optset, FLAG, 0, 'B', "methylcall", 
      "use simple methylation calling", NULL, NULL, &methylcall);
  manopt(&optset, FLAG, 0, 'Y', "methylbias", 
      "calculate methylation bias", NULL, NULL, &methylbias);





  unflagged = manopt_getopts(&optset, argc, argv);
  saveindex = manopt_isset(&optset, 'x', NULL);
  
  if(!(!manopt_isset(&optset, 'i', NULL) ^ !manopt_isset(&optset, 'x', NULL))) {
    manopt_help(&optset, "please give index filename using -i XOR -x option\n");
  } else if(unflagged->noofvalues > 1) { 
    manopt_help(&optset, "unknown argument(s)\n");
  }

  MSG("reading database sequences.\n"); 
  NFO("minsplitno set to %d\n", minsplitno);

  dbfilenames = manopt_getarg(&optset, 'd', "database");
  fasta = bl_fastxGetSet(space, dbfilenames->values, 
      dbfilenames->noofvalues, 1, 0, 0, 1);

  NFO("%d database sequences found.\n", fasta->noofseqs);
  MSG("reading query files.\n");

  queries = manopt_getarg(&optset, 'q', "query");
  if(queries->noofvalues > 30) {
    manopt_help(&optset, "currently no more than 30 query files allowed\n");
  }

  grouplist = manopt_getarg(&optset, 'g', "group");
  if(grouplist) {
    if(grouplist->noofvalues != queries->noofvalues) 
      manopt_help(&optset, "please provide a group name for each input file");

    groups = ALLOCMEMORY(space, NULL, Uint, grouplist->noofvalues);
    
    for(i=0; i < grouplist->noofvalues; i++) {
      groups[i] = atoi(grouplist->values[i]);
      if(groups[i] == 0)
        manopt_help(&optset, "please provide group numbers (int) > 0");
      if(groups[i] > queries->noofvalues)
        manopt_help(&optset, "please provide groupnumbers <= number of input files");
      NFO("found group number %d\n", groups[i]);
    }
  }

  if (methylcall || methylbias){
    if (queries->noofvalues > 1){
      manopt_help(&optset, "multiple query files are not supported for methylation analyses\n");
    }
  }

  if(saveindex) {
    indices = manopt_getarg(&optset, 'x', "generate");
  } else {
    indices = manopt_getarg(&optset, 'i', "index");
  }

  if(indices->noofvalues != queries->noofvalues) {
    manopt_help(&optset, "please provide an index file name for each query file\n");
  }

  ntcode  = getNTcodekey(space);
  files = ALLOCMEMORY(space, NULL, matchfile_t*, queries->noofvalues);

  for(i=0; i < queries->noofvalues; i++) {

    files[i] = ALLOCMEMORY(space, NULL, matchfile_t, 1);  
    files[i]->fmt = 0;
    files[i]->index = NULL;
    files[i]->filename = queries->values[i];

    prefixlen = bl_fileprefixlen(files[i]->filename);

    if(strncmp(&files[i]->filename[prefixlen], ".gz", 3) == 0 || 
        strncmp(&files[i]->filename[prefixlen], ".gzip", 5) == 0) {
      gzip = 1;
    }

    files[i]->gzip = gzip;

    if(saveindex) {

      bl_matchfileIndex(space, files[i], fasta);
      bl_matchfileWriteIndex(files[i]->index, indices->values[i]);  
      if(stats) { 
        getstats(space, files[i], fasta, mincover, maxcover, minfrac, entropyfilter, GEV, usenativequal); 
      }

      bl_matchfileWriteIndex(files[i]->index, indices->values[i]);  
      idx2 = bl_matchfileReadIndex(space, indices->values[i]);

      fprintf(stderr, "compare index (%p:%p):%d\n", 
          (void*)files[i]->index, (void *)idx2, 
          bl_compareIndices(files[i]->index, idx2));
      bl_matchfileDestructIndex(space, idx2); 
      FREEMEMORY(space, idx2);

    } else if(indices->values[i]) {
    
      MSG("reading index file\n");
      files[i]->index = bl_matchfileReadIndex(space, indices->values[i]);
    }
   
    /*
     *  typically stats should be present if stats
     *  failed however we are not allowed to set
     *  this one
     */
    if(files[i]->index->stats) { 
        if(files[i]->index->stats->mincover != mincover || 
            files[i]->index->stats->maxcover != maxcover || 
            files[i]->index->stats->minfrac != minfrac) {
          MSG("WARNING: you are resetting minfrac, mincover or maxcover to other values than used during sampling!");
        }
        files[i]->index->stats->mincover = mincover;
        files[i]->index->stats->maxcover = maxcover; 
        files[i]->index->stats->minfrac = minfrac;
        files[i]->index->stats->entropyfilter = entropyfilter;
        files[i]->index->stats->usenativequal = usenativequal;
        files[i]->index->stats->usegev = GEV;

    }

    //fprintf(stderr, "maxcover: %d\n",files[i]->index->stats->maxcover);

    /*
     * if stats are not present in any file
     * neither gev, dumpstats, nor call
     * can be called
     */
    if (!files[i]->index->stats &&
        (GEV || dumpstats || call || starts)){
      manopt_help(&optset, "please generate an index with stats (option -H) if the options -Z, -c, -S are used\n");
    }
  }


  if(manopt_isset(&optset, 'a', "annotation")) { 
    bedfilenames = manopt_getarg(&optset, 'a', "annotation");
    for(i=0; i < bedfilenames->noofvalues; i++) { 
  
      prefixlen = bl_fileprefixlen(bedfilenames->values[i]);
      if( strncmp(&bedfilenames->values[i][prefixlen], ".bed", 3) == 0 ||
        strncmp(&bedfilenames->values[i][prefixlen], ".BED", 3) == 0) {
        track = bl_BEDread(space, bedfilenames->values[i]);

        beddev = fopen("sorted.bed", "w");
        if(beddev  == NULL) {
          fprintf(stderr, "could not open file %s. Exit forced.", "sorted.bed");
          exit(-1);
        }

        bl_BEDwrite(track, beddev);
      
      } else if( strncmp(&bedfilenames->values[i][prefixlen], ".gff", 3) == 0 ||
        strncmp(&bedfilenames->values[i][prefixlen], ".GFF", 3) == 0) {
        track = bl_GFFread(space, bedfilenames->values[i]);
        bl_BEDwrite(track, beddev);
      
      } else {
        manopt_help(&optset, "please provide files with .GFF or .BED extension\n");
      }
    }
  }

  /*
   * TODO: stats for all files, what if stats failed!
   */
  
  if(gatk) {fprintf(stderr, "using GATK model\n"); gatkeval(space, files, groups, queries->noofvalues, fasta, maxframesize, usenativequal); }
  if(consensus){ fprintf(stderr, "calling consensus\n"); evalconsensus(space, files, groups, queries->noofvalues, fasta, maxframesize); }
  if(GEV)  { fprintf(stderr, "fitting GEV\n"); bl_matchfileFitGEV (NULL, files[0]->index->stats); }
  if(dumpstats) bl_matchfileDumpSampleStats(files[0]->index->stats);
  if(call) eval(space, files, groups, queries->noofvalues, fasta, maxframesize);
  if(simplegev) simpleeval(space, files, groups, queries->noofvalues, fasta, maxframesize, usenativequal);
  if(methylcall) callMethylSimple(space, files, groups, queries->noofvalues, fasta, maxframesize);
  if(methylbias) calcMethylBias(space, files, groups, queries->noofvalues, fasta, maxframesize);

  if(manopt_isset(&optset, 's', "splice") && 
      manopt_isset(&optset, 'q', "query")) {

      if(!splicebasename) { 
        splicebasename = bl_basename(files[0]->filename);
      }

      splicebasenamelen  = strlen(splicebasename);

      filename = ALLOCMEMORY(space, NULL, char, splicebasenamelen+7+4+1);
      sprintf(filename, "%s.splice.bed", splicebasename);

      transfilename = ALLOCMEMORY(space, NULL, char, splicebasenamelen+7+4+1);
      sprintf(transfilename, "%s.trans.bed", splicebasename);

      // not used: splice = 1;
      evalsplice(space, files, groups, queries->noofvalues, fasta, maxframesize, track, filename, transfilename, minsplitno);

      FREEMEMORY(space, filename);
      FREEMEMORY(space, transfilename);
  }

  
  if(browse) view(space, files, queries->noofvalues, fasta, track);
  if(starts) evalstarts(space, files, groups, queries->noofvalues, fasta, maxframesize, track);
  if(wiggle) writeexpression(space, files, groups, queries->noofvalues, fasta, track);

  bl_fastaDestruct(space, fasta);
  FREEMEMORY(space, fasta);
  
  if(files) {
    for(i=0; i < queries->noofvalues; i++) { 
      if (files[i]->index) {
        bl_matchfileDestructIndex(space, files[i]->index);
        FREEMEMORY(space, files[i]->index);
      }
      FREEMEMORY(space, files[i]);
    }
    FREEMEMORY(space, files);
  }

  if(groups) {
    FREEMEMORY(space, groups);
  }


  if(track) { 
    bl_annotationtrackDestruct(space, track);
    FREEMEMORY(space, track);
  }


  manopt_destructoptionset(&optset);
  manopt_destructarg(unflagged);
  FREEMEMORY(space, unflagged);

  FREEMEMORY(space, ntcode);

  return 0;
}

